////////////////////////////////////////////////////////////////////////////
//                           **** WAVPACK ****                            //
//                  Hybrid Lossless Wavefile Compressor                   //
//                Copyright (c) 1998 - 2026 David Bryant.                 //
//                          All Rights Reserved.                          //
//      Distributed under the BSD Software License (see license.txt)      //
////////////////////////////////////////////////////////////////////////////

// wavpack_version.h

#ifndef WAVPACK_VERSION_H
#define WAVPACK_VERSION_H

#define LIBWAVPACK_MAJOR 5
#define LIBWAVPACK_MINOR 9
#define LIBWAVPACK_MICRO 0
#define LIBWAVPACK_VERSION_STRING "5.9.0"

#endif
